# Полезные заметки
В данном проекте представлен код для развертки инфраструктуры описанной в первой таске из вики проекта.
Этот файл является небольшой сборкой полезных заметок при изучении terraform и AWS.

Содержание:
<div align="left">
<ul>
  <li>
    <a href="#aws-lambda-и-rust">AWS Lambda и Rust</a>
    <ul>
      <li>
        <a href="#как-написать-свою-лямбду-на-rust">
          Как написать свою лямбду на Rust
        </a>
      </li>
      <li>
        <a href="#rust-as-lambda">
          Rust as Lambda
        </a>
      </li>
      <li>
        <a href="#использование-стронних-сервисов-aws-в-лямбде">
          Использование стронних сервисов AWS в лямбде
        </a>
      </li>
      <li>
        <a href="#проверка-лямбды-локально">
          Проверка лямбды локально
        </a>
      </li>
    </ul>
  </li>
  <li>
    <a href="#некоторые-проблемы-с-terraform-и-их-решения">
      Некоторые проблемы с Terraform и их решения
    </a>
    <ul>
      <li>
        <a href="#логи-с-лямбды-не-пишутся-в-cloudwatch-там-какая-то-ошибка">
          Логи с лямбды не пишутся в cloudwatch. Там какая-то ошибка
        </a>
      </li>
      <li>
        <a href="#firehose-принимает-данные-но-лямбда-не-отрабатывает-или-в-s3-всегда-processing-failed">
          Firehose принимает данные, но лямбда не отрабатывает или в s3 всегда processing-failed
        </a>
      </li>
      <li>
        <a href="#при-сохранении-в-s3-появляется-пустая-папка">
          При сохранении в S3 появляется пустая папка
        </a>
      </li>
      <li>
        <a href="#немного-о-структуре-terraform-проектов">
          Немного о структуре terraform-проектов
          <ul>
            <li>
            <a href="#модули">Модули</a>
            </li>
            <li>
              <a href="#где-хранить-policy-и-roles">Где хранить policy и roles?</a>
            </li>
          </ul>
        </a>
      </li>
      <li>
        <a href="#sns-публикует-сообщения-но-sqs-пуст">
          Sns публикует сообщения, но sqs пуст
        </a>
      </li>
      <li>
        <a href="#триггер-не-указывается-в-lambda">
          Триггер не указывается в Lambda
        </a>
      </li>
      <li>
        <a href="#не-могу-переиспользовать-бакет-s3">
          Не могу переиспользовать бакет S3
        </a>
      </li>
      <li>
        <a href="#что-то-я-не-понял-а-как-работать-с-asg-и-ec2">
          Что-то я не понял, как работать с Auto scaling group и решать задачи
        </a>
      </li>
    </ul>
  </li>
  <li>
    <a href="#три-закона-робототехники-айзека-азимова">
      Три закона робототехники Айзека Азимова
    </a>
  </li>
</ul>
</div>

# AWS Lambda и Rust
<div align="center">
  <img src="./images/lambda-rust.png">
</div>
Тут описано как написать свою лямбду на функциональном языке [Rust](https://www.rust-lang.org/) с небольшими поправками относительно оффициального туториала.
В примере есть копия кода лямбды с оффициального туториала Amazon, только с поддержкой асинхронности.

# Как написать свою лямбду на Rust
Для создания лямбды на раст Amazon советует использовать и оффициальную библиотеку. Именно её мы и будем использовать,
вот только с небольшим условием. В оффициальном репозитории пакетов для раста ([crates](https://crates.io)) данный пакет обновлялся год назад
и не поддерживает асинхронность. Поэтому тянуть будем напрямую с git. Для этого добавим в Cargo.toml зависимость через git.
```toml
[dependencies.lambda]
git = "https://github.com/awslabs/aws-lambda-rust-runtime"
```
Теперь у нас есть последняя версия нужной библиотеки.
Далее Напишем следующее в main:
```rust
#[macro_use]
extern crate log;
#[macro_use]
extern crate serde;

use std::error::Error;
use tokio::runtime::Runtime;

type LambdaError = Box<dyn Error + Send + Sync + 'static>;

#[derive(Deserialize, Debug, Clone)]
struct CustomEvent {
    #[serde(rename = "firstName")]
    first_name: String,
}

#[derive(Serialize, Debug, Clone)]
struct CustomOutput {
    message: String,
}

/// This function starting point of the binary. It points the lambda to execution function.
/// Real lambda processor located in `processors` module.
fn main() -> Result<(), LambdaError> {
    simple_logger::init_with_level(log::Level::Info)?;
    let mut rt = Runtime::new().expect("can't start Tokio runtime");
    let handler = lambda::handler_fn(process);
    rt.block_on(lambda::run(handler))?;
    Ok(())
}

/// Main lambda processor.
/// All new events will be deserialized and cast to the "CustomEvent".
/// Make sure that your "CustomEvent" is matching AWS generated event.
pub async fn process(e: CustomEvent) -> Result<CustomOutput, LambdaError> {
    info!("Got event: {:#?}", e);
    if e.first_name == "" {
        return Err(c.new_error("Empty first name"));
    }
    Ok(CustomOutput{
        message: String::from("This guy has name")
    })
}
```

Данный кусок кода - точная копия примера из оффициальной статьи Amazon, только хендлер является асинхронным.
Это потребуется для использования сторонних сервисов Амазона в лямбдах.

## Rust as Lambda
Немного о расте как о среде для запуска лямбд в AWS.
Раст является компилируемым языком и реального рантайма под него в AWS нет,
и именно поэтому нужно использовать 
рантайм [Provided](https://docs.aws.amazon.com/lambda/latest/dg/runtimes-custom.html);
А для моделей AWS, которыми обмениваются сервисы слеует подтянуть библиотеку `aws_lambda_events`, однако требуется строго версия 0.2.5!
Потому что в более новых версиях библиотеки у Firehose появилась обязательное поле `kinesisRecordMetadata`, которого в реальности зачастую нет. 

После этого, даже указав рантайм, надо правильно слинковать бинарный файл, а для этого надо:
Добавим в Cargo.toml секцию для указания имени результируюшего бинарного файла:
```toml
[[bin]]
name = "bootstrap"
path = "src/main.rs"
```

После этого добавим в rustup интересющий нас таргет и перейдем на nightly toolchain.
Для этого в терминале вводим:
```bash
> rustup toolchain add nightly
> rustup target add x86_64-unknown-linux-musl
```
Теперь мы можем копилировать проект под musl-gcc и собрать его в zip.
```bash
> cargo build --release --target x86_64-unknown-linux-musl
> zip -j "lambda.zip" target/x86_64-unknown-linux-musl/release/bootstrap
```

Итогом компиляции должен быть пакет без динамических библиотек. 
Поподрбнее можно почитать [тут](https://doc.rust-lang.org/1.9.0/book/advanced-linking.html#static-linking);

## Использование стронних сервисов AWS в лямбде
 Для использования sns и какого-либо другого сервиса внутри лямбды потребуется пакеты 
`rusoto_core` и `rusoto_<название сервиса>`, например `rusoto_sns`. **Обязательно надо включить "rustls" в фитчи всех
 `rusoto_` пакетов, иначе кросс-компиляция не сработает и пакет будет зависить от системного libssl
  (Такая лямбда не запустится в AWS)**;

Все библиотеки группы `rustoto_` являются асинхронными, поэтому можно либо испольовать `async-await` синтаксис,
либо блокировать тред вручную через `futures`. Первый авриант быстрее, так как std раста теперь поддреживает 
асинхронность "Из коробки", да и асинхронность никогда не помешает.


## Проверка лямбды локально
Для проверки лямбд на своей машине есть специальные образы докера, позволяющее эмулировать среду выполнения
лямбды на серверах амазона. 

Описание работы и прочие полезности можно найти в [этом репозитории](https://github.com/lambci/docker-lambda)

Примеры интеграционного тестирования можно видеть в Makefile (таргет integration_test) в проектах лямбды ссылки ниже:
* [тык](./task.1/lambdas/firehose_lambda_function/Makefile);
* [тык тык](./task.1/lambdas/sns_lambda_function/Makefile).

# Некоторые проблемы с Terraform и их решения

<div align="center">
  <img src="./images/terraform_troubles.png">
</div>

Тут описаны проблемы, с которыми я столкнулся при выполнении заданий и то, как их решить.
В основном проблемы довольно простые и распространенные. Но лучше уж где-нибудь записать на будущее.

## Логи с лямбды не пишутся в cloudwatch. Там какая-то ошибка
Если при проверке логов встретилась подобная ошибка:

![error_screenshot](./images/error_lambda_cloudwatch.png)

То следует поменять названия log-group и проверить права лямбды.
Название log-group для лямбды должно быть "`/aws/lambda/${aws_lambda_function.my_lambda.function_name}`", 
где `aws_lambda_function.my_lambda` -  путь до вашей существующей лямбды в конфигурации terraform.

## Firehose принимает данные, но лямбда не отрабатывает или в s3 всегда processing-failed
Скорее всего у вас недстаточно прав на запуск лямбды.
Перепроверьте, что у firehose есть все необходимые права.
Для первого задания нужны права на запуск лямбды, выгрузку логов и возможность записи в s3.

Вот конфигурация:
```terraform
data "aws_iam_policy_document" "fire_extra_policy_doc" {
  # allow to write logs
  statement {
    effect = "Allow"
    actions = [
      "logs:PutLogEvents"
    ]
    resources = [
      aws_cloudwatch_log_stream.kinesis_stream.arn
    ]
  }

  # allow to write to S3
  statement {
    effect = "Allow"
    actions = [
      "s3:AbortMultipartUpload",
      "s3:GetBucketLocation",
      "s3:GetObject",
      "s3:ListBucket",
      "s3:ListBucketMultipartUploads",
      "s3:PutObject"
    ]
    resources = [
      aws_s3_bucket.bucket.arn,
      "${aws_s3_bucket.bucket.arn}/*"
    ]
  }

  # allow to run transforming lambda
  statement {
    effect = "Allow"
    actions = [
      "lambda:InvokeFunction",
      "lambda:GetFunctionConfiguration"
    ]
    resources = [
      "${aws_lambda_function.firehorse_lambda.arn}:*"
    ]
  }
}
```

Потом следует приатачить права к вашей существующей роли firehose.

```terraform
resource "aws_iam_policy" "firehose_extra_policy" {
  name = "firehose-extra-permissions-policy"
  policy = data.aws_iam_policy_document.fire_extra_policy_doc.json
}


resource "aws_iam_role_policy_attachment" "firehose_iam_role_policy_attachment" {
  role       = aws_iam_role.firehose_role.name
  policy_arn = aws_iam_policy.firehose_extra_policy.arn
}
```
Тут `aws_iam_role.firehose_role` - путь до существующей ролим вашего firehose.
Если в AWS-S3 всё равно какие-то странные файлы в папке `processing-failed`, 
то скачивайте эти файлы и открывайте как обычный текстовый документ, 
там будет довольно точное и подробное описание проблемы.

## При сохранении в S3 появляется пустая папка
Уберите '/' в начале пути.
Например:

`/UNAUTHORIZED/{}/{}/{}.txt` - создаст путь `s3://   /UNAUTHORIZED/...`

`UNAUTHORIZED/{}/{}/{}.txt` - создаст путь `s3://UNAUTHORIZED/...` (Праильный вариант)


## Немного о структуре terraform-проектов
### Модули
Довольно удобная фитча тераформа - модули. В рамках данного проекта это не особо нужная штука, 
но всё же стоит посмотреть. Это позволяет переиспользовать какие-то части инфраструктуры с другими переменными.
Так, например, легко можно разделать прод, тест и какую угодно конфигурацию в однном проекте, а не плодить кучу проектов.

### Где хранить policy и roles?
Так как terraform является довольно свободной, в плане конфигурации, штукой то роли, да и вообще любые куски инфраструктуры,
можно хранить как угодно. Я пришел к тому, что хранить в одном файле роли, их полиси_доки и сами объекты инфраструктуры, где эти полиси и 
роли применяются куда удобнее, чем отдельный модуль с полиси и отдельный с ролями. За это озарение спасибо [Peter Zinin](skype:live:8422?chat).

С одной стороны подобная свобода в написании проекта это удобно. Но когда проект разрастается, то без должной структуры проект превращается в кашу.

Чтобы этого избежать, лучше всего логически разделять отдельные части инфраструктуры и писать в отдельных файлах всё необходимое для конкретного элемента, таким образом, чтобы минимизировать линки между файлами.
 
Как по мне, так писать в одном файле вообще не лучшая затея, так что, даже если проект небольшой, то 
лучше всё равно разделять его на какие-то части.

## Sns публикует сообщения, но sqs пуст
Скорее всего у вас не выданы права на публикацию сообщений в очередь.
Хоть и показано, что `SQS` привязан к топику `SNS` требуется выдача определенных прав.
Добавьте в конфигурацию террафаорм следующее:

```terraform

# Добавление прав публикации к очереди.
resource "aws_sqs_queue_policy" "sqs_publish_policy" {
  queue_url = aws_sqs_queue.<your_sqs_queue>.id
  policy    = data.aws_iam_policy_document.sqs_send_message_policy_doc.json
}

# Описание политики публикации
# Тут указано условие, что публикация возможна только из объявленного sns.
data "aws_iam_policy_document" "sqs_send_message_policy_doc" {
  statement {
    effect = "Allow"
    actions = [
      "sqs:SendMessage"
    ]
    principals {
      type = "AWS"
      identifiers = [
        "*"
      ]
    }
    resources = [
      aws_sqs_queue.<your_sqs_queue>.arn
    ]
    condition {
      test     = "ArnEquals"
      variable = "aws:SourceArn"
      values = [
        aws_sns_topic.<your_sns_topic>.arn
      ]
    }
  }
}
```

## Триггер не указывается в Lambda
Скорее всего у вас неверно выданы права на выполнение лямбды.

Попробуйте поменять `aws_iam_policy_document`, содержащий `actions: [lambda:InvokeFunction]` на 
следующий кусок кода:
```terraform

# Подставьте сюда свои названия лямбды и топика.
resource "aws_lambda_permission" "sns_allow_lambda_permission" {
  action        = "lambda:InvokeFunction"
  function_name = aws_lambda_function.<your_lambda>.function_name
  principal     = "sns.amazonaws.com"
  source_arn    = aws_sns_topic.<your_sns_topic>.arn
}
```
В консоле AWS лямбды должен появится sns как триггер. Выглядит это так:

<div align="center">
  <img src="./images/lambda_sns_trigger.png">
</div>

## Не могу переиспользовать бакет S3

Для переиспользования уже существующего бакета s3 или любого другого объекта инфраструктуры, перед 
`terraform apply` следует вручную обновить информацию о его состоянии в `terraform.tfstate`.
Для этого выполните команду в терминале:
```bash
#             Путь до объекта в terraform конфигах   Уникальный идентификатор в aws.
terraform import module.services.aws_s3_bucket.bucket "bucket-name"
```
Как импортировать другие объекты можно прочитать в доках терраформ в разделе Import.

## Что-то я не понял, а как работать с ASG и EC2
Auto scaling group штука довольно необычная и очень удобная, с одной стороны. C другой стороны, имеет
 некоторые моменты, которые стоит прояснить прежде чем работать с ним.

Давайте примерно разберем архитектуру ASG в данном проекте.
На картинке ниже представлена архитектура ASG со второго задания с некоторыми допущениями, которые мы разберем позднее.
<div align="center">
  <img src="./images/asg-architecture.png">
</div>

 ### Scale up and Scale down policies
 
 Начнем с того, что `Auto Scaling Group` не должен сам выбирать, когда добавлять новый `EC2 instance` и когда его останавливать, он просто умеен это делать. Для отслеживания состояний нам потребуется `CloudWatch` и его `Alarms`. Работают они крайне тупо, но эффективно. На диаграмме сверху я пометил связь между алармой и SQS как 
 `Observing metrics` потому что именно он будет нашим тригером. В данном случае `CloudWatch` через некоторые промежутки времени запращивает у `SQS` какие-нибудь метрики и передает их в свои алармы. Идея в том, что у `Alarm` есть некоторый триггер и действие, которое произойдет, когда условие тригера сработает. 
 
 Вот пример конфигурации Alarm относительно количества сообщений в SQS:
  ```terraform
  resource "aws_cloudwatch_metric_alarm" "sqs-asg-metric-alarm" {
    alarm_name          = "SQS-message-alarm"
    comparison_operator = "GreaterThanOrEqualToThreshold"
    evaluation_periods  = "1"
    metric_name         = "ApproximateNumberOfMessagesVisible"
    namespace           = "AWS/SQS"
    period              = "60"
    statistic           = "SampleCount"
    threshold           = "5"

    dimensions = {
      QueueName = aws_sqs_queue.s3rius_sqs_queue.name
    }

    alarm_description = "This metric monitors SQS messages and fire ec2 from asg"
    alarm_actions = [
      aws_autoscaling_policy.scale_up_policy.arn
    ]
  }
  ```

  Если одновременно в очереди висит больше 5 сообщений то вызывается `aws_autoscaling_policy.scale_up_policy`, где описано добавления одного инстанса.
  Также в проекте есть второй `aws_autoscaling_policy`, который дропает `EC2`, когда инстанс 4 минуты подряд
  простаивает с загрузкой CPU ниже 5%. 
  <div align="center">
    <img src="./images/asg-meme.png">
  </div>
  Замечу, что `action` у аларма - это просто arn какого-нибудь aws_autoscaling_policy. В данном случае полиси просто триггернется и сам решит что делать. Вообще в alarm можно встраивать не только тупые стандартные проверочки, но и целые вычисления.

  ### Я поднимаю EC2, а как обрабатывать сообщения SQS?
  Очень просто. Сам EC2 - это просто мелкий сервер, который работает как твой локальный комп. У него есть user-commands, которые выполняются при запуске. Просто прокинь туда запуск своего приложения на любом языке, которое будет слушать SQS и обрабатывать сообщения в вечном лупе. Если что, `AutoSclaingGroup` сам дропнет инстанс, который слишком долго работает ничего не обрабатывая.

  Бтв, если не настроил `AutoScalingPolicies`, рискуешь влететь на монетки. Не очень большие, но всё равно не очень приятно. Также замечу, что бесплатный тип инстанса `EC2` - это `t2.micro`, а **НЕ** `t2.nano`.

  ### Я понял, как писать user-commands, а как загрузить приложение на EC2?
  Скомпиль локально или просто заверни в zip-файл и через `terraform` загрузи в какой-нибудь `s3-bucket`.
  Вот пример такой загрузки:
  ```terraform
  // bucket - ведро, куда подгрузить файл приложения;
  // key    - путь в s3 до файла;
  // source - путь до локального файла. 
  resource "aws_s3_bucket_object" "application_data" {
    bucket = aws_s3_bucket.asg_meta_bucket.bucket
    key    = var.ec2_application_s3_key
    source = var.ec2_application_bin
  }
  ```
  После развертывания инфраструктуры у тебя будет объект в s3-bucket с самого начала. Никакой ручной загрузки.
  А в самом инстансе внутри user-command скачай файл через aws-cli и разархивируй его.
  ```bash
  aws s3 cp "s3://${meta_bucket}/${application_key}" "application.zip"
  unzip application.zip
  ```
  А вот дальше запускай как знаешь.

  ### Я разархиаировал приложение, но для запуска мне нужны сторонние тулы/языки, которых нет
  EC2 - обычная `unix`'овая машинка. А у этих ребят есть пакетные менеджеры и прочие штуки, чтобы 
  удобно что-нибудь установить. У AmazonLinux, который ты, скорее всего выбрал, стоит yum (Это как у fedora, если шаришь, если нет, то зырк [сюды](https://habr.com/ru/post/301292/)).

  А вот для работы `YUM` нужен выход в тырнетики, чтобы он мог всё скачать. Для этого решений парочка.
  Например:
  * Развернуть `VPC`;
  * Настроить `aws_internet_gateway` и подключить его к `VPC`.
  * Создать `aws_subnet`, с `cidr_block = "10.1.1.0/24"`
  * Подрубить балансировщик и все инстансы к этой сетке.

  Но это охрененно сложно, как я думаю. Поэтому предлагаю вариант, который
  почему-то не освещается на первых страницах гугла. (Чертовски подозрительно, кстати);

  Создадим штуку, которая будет пускать траффик куда угодно по любому порту.
  в `aws_security_group` `egress` - это исходящий траффик, `ingress` - входящий.
  ```terraform
  resource "aws_security_group" "outbound_traffic_security" {
  name = "outbound-ec2"

  # Allow all outbound
  egress {
    from_port = 0
    to_port   = 0
    protocol  = "-1"
    cidr_blocks = [
      "0.0.0.0/0"
    ]
  }

  # Inbound HTTP from anywhere
  ingress {
    from_port = var.ec2_port
    to_port   = var.ec2_port
    protocol  = "tcp"
    cidr_blocks = [
      "0.0.0.0/0"
    ]
  }
}
```

А потом в своем `aws_launch_configuration` добавь:
```terraform
  security_groups = [
    aws_security_group.inbound_traffic_security.id,
    aws_security_group.outbound_traffic_security.id
  ]
```

Теперь можно смело скачивать хоть весь интернет. Но если хочется чего-то конкретного поставить. Используй пакетный менеджер или `wget`. Например так:
```bash
yum update -y
yum install <package>
```
Чтоб узнать есть ли какой-нибудь пакет в `AmazonLinux` репозиториях посмотри [полный список доступных пакетов](https://aws.amazon.com/ru/amazon-linux-ami/2018-03-packages/).

### Как отправлять логи ec2 в CloudWatch?
  <div align="center">
    <img src="./images/ec2-logging-meme.jpg">
  </div>

  Логгирование в stdout на ec2 не вариант, если не знать как. Так-то все логи пишутся в systemlog, который можно почитать после выключения инстанса. Но это неудобно. Поэтому предлагаю настроить user-command так, чтобы весь лог из stdout записывался в кастомный файл.
  ```bash
  #!/bin/bash
  LOG_FILE="/var/log/my_logs.log"

  function main() {
    # Your awesome logic
  }

exec > >(tee "$LOG_FILE" | logger -t user-data -s 2>/dev/console) 2>&1
    echo "################# INSTANCE STARTUP #################"
    main
  ```
  Потом спокойно подключаемся по ssh и в реальном времени читаем логи с помощью:
  ```bash
  tail -f "/var/log/my_logs.log"
  ```
  Также есть вариант подключить CloudWatch агента. С этим я особо не разбирался, но если интересно, то 
  [смотреть сюда](https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/install-CloudWatch-Agent-commandline-fleet.html).

# Три закона робототехники Айзека Азимова
  * Робот не может причинить вред человеку или своим бездействием допустить, чтобы человеку был причинён вред.
  * Робот должен повиноваться всем приказам, которые даёт человек, кроме тех случаев, когда эти приказы противоречат Первому Закону.
  * Робот должен заботиться о своей безопасности в той мере, в которой это не противоречит Первому или Второму Законам.