use rand::{
    distributions::{Distribution, Standard},
    Rng,
};
use std::time::{SystemTime, UNIX_EPOCH};

/// Possible station types.
/// The lambda will process data only if
/// the station type is `AUTHORIZED`
#[derive(Serialize, Deserialize, Clone, Debug)]
pub enum StationTypes {
    AUTHORIZED,
    UNAUTHORIZED,
    UNKNOWN,
}

impl Distribution<StationTypes> for Standard {
    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> StationTypes {
        match rng.gen_range(0, 3) {
            0 => StationTypes::AUTHORIZED,
            1 => StationTypes::UNAUTHORIZED,
            _ => StationTypes::UNKNOWN,
        }
    }
}

/// AWS lambda input data.
/// Input will be mapped from json to this type using `serde_json`.
#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct WeatherDataInput {
    #[serde(rename = "Timestamp")]
    pub timestamp: usize,
    #[serde(rename = "Coords")]
    pub coords: usize,
    #[serde(rename = "Temperature")]
    pub temperature: i64,
    #[serde(rename = "Humidity")]
    pub humidity: i64,
    #[serde(rename = "Pressure")]
    pub pressure: i64,
    #[serde(rename = "StationType")]
    pub station_type: StationTypes,
    #[serde(rename = "IpAddress")]
    pub ip_addr: String,
}

impl Distribution<WeatherDataInput> for Standard {
    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> WeatherDataInput {
        let start = SystemTime::now();
        let since_the_epoch = start.duration_since(UNIX_EPOCH)
            .expect("Time went backwards");
        WeatherDataInput {
            timestamp: since_the_epoch.as_millis() as usize,
            coords: rng.gen::<usize>() % 10000,
            temperature: rng.gen::<i64>() % 100,
            humidity: (rng.gen::<i64>() % 100).abs(),
            pressure: (rng.gen::<i64>() % 400).abs(),
            station_type: rng.gen(),
            ip_addr: String::from("192.168.1.1"),
        }
    }
}


#[must_use]
pub fn get_firehose_input_example() -> String {
    include_str!("data_examples/weather_data_input.json").to_string()
}


#[cfg(test)]
mod tests {
    use crate::WeatherDataInput;

    #[test]
    pub fn test_input_parsing() {
        let data = include_str!("data_examples/weather_data_input.json");
        serde_json::from_str::<WeatherDataInput>(data).unwrap();
    }
}
