use crate::locator::{locate_city, locate_country};
use crate::WeatherDataInput;

/// AWS lambda return type.
/// All fields are copied from the input, except
/// `city` and `country`. The generated according to coords field.
#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct WeatherDataOutput {
    pub timestamp: usize,
    pub coords: usize,
    pub temperature: i64,
    pub humidity: i64,
    pub pressure: i64,
    pub city: String,
    pub country: String,
    pub ip_addr: String,
}

impl From<WeatherDataInput> for WeatherDataOutput {
    fn from(input: WeatherDataInput) -> Self {
        debug!("Parsing");
        Self {
            timestamp: input.timestamp,
            coords: input.coords,
            temperature: input.temperature,
            humidity: input.humidity,
            pressure: input.pressure,
            ip_addr: input.ip_addr,
            city: locate_city(input.coords),
            country: locate_country(input.coords),
        }
    }
}

#[cfg(feature = "parsing_parquet")]
pub mod parse_parquet {
    use super::WeatherDataOutput;
    use tokio::fs::File;
    use std::path::Path;
    use parquet::file::reader::{SerializedFileReader, FileReader};
    use parquet::record::RowAccessor;
    use std::error::Error;

    impl WeatherDataOutput {
        pub async fn from_parquet_file(file_name: String) -> Result<Vec<Self>, Box<dyn Error + Send + Sync + 'static>> {
            let parquet_file = File::open(&Path::new(file_name.as_str()))
                .await?
                .into_std()
                .await;
            let reader = SerializedFileReader::new(parquet_file).unwrap();
            let iter = reader.get_row_iter(None)?;
            let mut results = Vec::new();
            for record in iter {
                results.push(Self {
                    timestamp: record.get_long(0)? as usize,
                    coords: record.get_long(1)? as usize,
                    temperature: record.get_long(2)?,
                    humidity: record.get_long(3)?,
                    pressure: record.get_long(4)?,
                    city: record.get_string(5)?.clone(),
                    country: record.get_string(6)?.clone(),
                    ip_addr: record.get_string(7)?.clone(),
                });
            }
            Ok(results)
        }
    }
}

#[must_use]
pub fn get_firehose_output_example() -> String {
    include_str!("data_examples/weather_data_output.json").to_string()
}

#[cfg(test)]
mod tests {
    use crate::{StationTypes, WeatherDataInput};
    use super::WeatherDataOutput;

    #[test]
    pub fn test_mapping_inputs() {
        let input = WeatherDataInput {
            timestamp: 12_393_994,
            coords: 3003,
            temperature: 12334,
            humidity: 1231,
            pressure: 3000,
            station_type: StationTypes::AUTHORIZED,
            ip_addr: String::from("192.168.1.1"),
        };
        WeatherDataOutput::from(input);
    }
}
