resource "aws_kinesis_firehose_delivery_stream" "s3rius_fire_stream" {
  name        = "kinesis-s3rius-fire-stream"
  destination = "extended_s3"

  # Configuration for firehose to process inputs
  # with lambda and serilize it in Parquet with AWS-Glue
  extended_s3_configuration {
    role_arn            = aws_iam_role.firehose_role.arn
    bucket_arn          = aws_s3_bucket.bucket.arn
    prefix              = "AUTHORIZED/"
    error_output_prefix = "FAILED/"
    buffer_size         = 128
    buffer_interval     = 60

    cloudwatch_logging_options {
      enabled         = true
      log_group_name  = aws_cloudwatch_log_group.kinesis_log_group.name
      log_stream_name = aws_cloudwatch_log_stream.kinesis_stream.name
    }

    processing_configuration {
      enabled = "true"

      processors {
        type = "Lambda"

        parameters {
          parameter_name  = "LambdaArn"
          parameter_value = "${aws_lambda_function.firehorse_lambda.arn}:$LATEST"
        }
      }
    }
    data_format_conversion_configuration {
      input_format_configuration {
        deserializer {
          open_x_json_ser_de {}
        }
      }
      output_format_configuration {
        serializer {
          parquet_ser_de {
            compression = "GZIP"
          }
        }
      }
      schema_configuration {
        database_name = aws_glue_catalog_database.glue_database.name
        role_arn      = aws_iam_role.firehose_role.arn
        table_name    = aws_glue_catalog_table.glue_table.name
      }
    }
  }
}


resource "aws_iam_role" "firehose_role" {
  name               = var.firehose_role_name
  assume_role_policy = data.aws_iam_policy_document.firehose_role_policy_doc.json
}

resource "aws_iam_policy" "firehose_extra_policy" {
  name   = "firehose-extra-permissions-policy"
  policy = data.aws_iam_policy_document.fire_extra_policy_doc.json
}


resource "aws_iam_role_policy_attachment" "firehose_iam_role_policy_attachment" {
  role       = aws_iam_role.firehose_role.name
  policy_arn = aws_iam_policy.firehose_extra_policy.arn
}


data "aws_iam_policy_document" "firehose_role_policy_doc" {
  version = "2012-10-17"
  statement {
    effect = "Allow"
    actions = [
      "sts:AssumeRole"
    ]

    principals {
      type = "Service"
      identifiers = [
        "firehose.amazonaws.com"
      ]
    }
  }
}

data "aws_iam_policy_document" "fire_extra_policy_doc" {
  version = "2012-10-17"
  # allow to write logs
  statement {
    effect = "Allow"
    actions = [
      "logs:PutLogEvents"
    ]
    resources = [
      aws_cloudwatch_log_stream.kinesis_stream.arn
    ]
  }

  # allow to write to S3
  statement {
    effect = "Allow"
    actions = [
      "s3:AbortMultipartUpload",
      "s3:GetBucketLocation",
      "s3:GetObject",
      "s3:ListBucket",
      "s3:ListBucketMultipartUploads",
      "s3:PutObject"
    ]
    resources = [
      aws_s3_bucket.bucket.arn,
      "${aws_s3_bucket.bucket.arn}/*"
    ]
  }

  # allow to run transforming lambda
  statement {
    effect = "Allow"
    actions = [
      "lambda:InvokeFunction",
      "lambda:GetFunctionConfiguration"
    ]
    resources = [
      "${aws_lambda_function.firehorse_lambda.arn}:*"
    ]
  }
}