#[macro_use]
extern crate log;

use async_std::task;
use env_logger::Env;
use lambda_models::WeatherDataInput;
use rusoto_firehose::{
    KinesisFirehose, KinesisFirehoseClient, PutRecordInput, PutRecordOutput, Record,
};
use rusoto_signature::Region;
use std::error::Error;
use std::time::Duration;
use std::str::FromStr;

#[tokio::main]
pub async fn main() {
    dotenv::dotenv().ok();
    let mut messages_quantity = 10;
    if let Some(quantity) = std::env::args().nth(1) {
        println!("Desired messages quantity {}", quantity);
        messages_quantity = i32::from_str(quantity.as_str()).expect("Not a number");
    }
    env_logger::init_from_env(Env::default().default_filter_or("debug"));
    let client = rusoto_firehose::KinesisFirehoseClient::new(Region::EuCentral1);
    let delivery_stream_name = String::from("kinesis-s3rius-fire-stream");
    for _ in 0..messages_quantity {
        let message = publish_message(&client, delivery_stream_name.clone()).await;
        match message {
            Ok(out) => {
                info!("Record successfully published");
                debug!("RecordId: {}", out.record_id);
                debug!("Encrypted: {:#?}", out.encrypted);
            }
            Err(err) => {
                error!("Message wasn't sent");
                error!("{:#?}", &err.to_string())
            }
        }
        info!("Sleeping");
        task::sleep(Duration::from_secs(1)).await;
    }
}

async fn publish_message(
    client: &KinesisFirehoseClient,
    delivery_stream_name: String,
) -> Result<PutRecordOutput, Box<dyn Error>> {
    let weather_data: WeatherDataInput = rand::random();
    let data_json = serde_json::to_string_pretty(&weather_data)?;
    info!("Sending data:");
    info!("{}", data_json);
    let out = client
        .put_record(PutRecordInput {
            delivery_stream_name,
            record: Record {
                data: bytes::Bytes::from(data_json),
            },
        })
        .await?;
    Ok(out)
}
