use crate::LambdaError;
use aws_lambda_events::encodings::Base64Data;
use aws_lambda_events::event::firehose::{
    KinesisFirehoseEvent, KinesisFirehoseEventRecord, KinesisFirehoseResponse,
    KinesisFirehoseResponseRecord,
};
use lambda_models::{StationTypes, WeatherDataInput, WeatherDataOutput};
use rusoto_core::Region;
use rusoto_sns::{PublishInput, Sns, SnsClient};
use serde_json::{Map, Value};
use std::io::ErrorKind;
use std::str::FromStr;

/// Main data processing point.
///
/// This function proceed firehose input buffers
/// I use `KinesisFirehoseEvent` here, because it's
/// common `FireHose` input format representation
/// Check more in official AWS documentation.
///
/// # Errors
///
/// Will return `Err` if any problem occur during processing.
pub async fn process(event: KinesisFirehoseEvent) -> Result<KinesisFirehoseResponse, LambdaError> {
    info!("Lambda ignition");
    let mut processed_events = Vec::new();
    for record in event.records {
        match process_input(record).await {
            Ok(response) => {
                info!("Data processed successfully.");
                processed_events.push(response)
            }
            Err(err) => warn!("Can't process data. Cause: {}", err.to_string()),
        }
    }
    Ok(KinesisFirehoseResponse {
        records: processed_events,
    })
}

/// Process actual input from Kinesis firehose.
/// This function takes one single record,
/// trying to parse the data and do actual processing
///
/// If the input station type is `UNKNOWN` or the data json is invalid,
/// then response data will be marked as `ProcessingFailed` and returned ro firehose.
///
/// If the stations type is `UNAUTHORIZED`,
/// then it will locate the city and send the data to SNS.
/// Also the transformation result will be marked as Dropped.
///
/// If the stations type is `AUTHORIZED`,
/// then it will locate the city and the transformation result will be marked as Ok.
///
/// # Errors
///
/// * Will return `Err` if bytes from input record can't be represented as
/// normal `UTF-8` string.
/// * If Weather data cannot be formatted as json.
/// * If publishing to SNS failed.
async fn process_input(
    record: KinesisFirehoseEventRecord,
) -> Result<KinesisFirehoseResponseRecord, LambdaError> {
    let processing_failed = KinesisFirehoseResponseRecord {
        record_id: record.record_id.clone(),
        result: Some(String::from("ProcessingFailed")),
        data: record.data.clone(),
    };
    let weather_json = String::from_utf8(record.data.to_vec());
    if weather_json.is_err() {
        return Ok(processing_failed);
    }
    let weather_json = weather_json.unwrap();

    info!("Received data: {}", weather_json);
    let weather_data = serde_json::from_str::<WeatherDataInput>(weather_json.as_str());

    if let Err(parse_err) = weather_data {
        error!(
            "Found error while parsing JSON data: {}",
            parse_err.to_string()
        );
        return Ok(processing_failed);
    }
    let weather_data = weather_data.unwrap();
    match weather_data.station_type {
        StationTypes::AUTHORIZED => {
            let authorized_output = WeatherDataOutput::from(weather_data.clone());
            let json_output = serde_json::to_vec(&authorized_output)?;
            debug!("Data is OK and AUTHORIZED");
            Ok(KinesisFirehoseResponseRecord {
                record_id: record.record_id,
                result: Some(String::from("Ok")),
                data: Base64Data(json_output),
            })
        }
        StationTypes::UNAUTHORIZED => {
            let authorized_output = WeatherDataOutput::from(weather_data.clone());
            if publish_to_sns(authorized_output.clone()).await.is_err() {
                warn!("Can't publish message to sns");
                return Ok(processing_failed);
            }

            let json_output = serde_json::to_vec(&authorized_output)?;
            debug!("Data is OK but UNAUTHORIZED");
            Ok(KinesisFirehoseResponseRecord {
                record_id: record.record_id,
                result: Some(String::from("Dropped")),
                data: Base64Data(json_output),
            })
        }
        StationTypes::UNKNOWN => Ok(processing_failed),
    }
}

/// Function to perform SNS publishing.
/// It will publish generated object into sns
///
/// `AWS_REGION` is variable provided by AWS lambda environment.
/// `SNS_TOPIC_ARN` is injected by terraform.
/// # Errors
/// Function will stop execution in several cases:
/// * The `AWS_REGION` variable not found;
/// * `AWS_REGION` env variable is incorrect;
/// * The `SNS_TOPIC_ARN` variable not found;
/// * Can't establish connection to SNS;
/// * Can't build `JSON` from `WeatherDataOutput`;
/// * Publish to topic failed by any reason.
pub async fn publish_to_sns(output: WeatherDataOutput) -> Result<(), LambdaError> {
    info!("Publishing to SNS");
    let current_region = std::env::var("AWS_REGION")?;
    let sns_client = SnsClient::new(Region::from_str(current_region.as_str()).map_err(|err| {
        std::io::Error::new(
            ErrorKind::Other,
            format!("Wrong region. {}", err.to_string()),
        )
    })?);
    debug!("Created SNS client");
    let message_string = serde_json::to_string(&output)?;
    let mut obj = Map::default();
    // Adding default field for SNS message.
    obj.insert(String::from("default"), Value::String(message_string));
    let message = serde_json::to_string_pretty(&obj)?;
    info!("Built message");
    let mut sns_message = PublishInput::default();
    sns_message.message = message;
    sns_message.topic_arn = Some(std::env::var("SNS_TOPIC_ARN")?);
    sns_message.message_structure = Some(String::from("json"));
    sns_client.publish(sns_message).await?;
    info!("Message successfully published to SNS");
    Ok(())
}

#[cfg(test)]
mod tests {
    use crate::processors::process_input;
    use aws_lambda_events::encodings::{Base64Data, MillisecondTimestamp};
    use aws_lambda_events::event::firehose::KinesisFirehoseEventRecord;
    use chrono::{NaiveDateTime, Utc};
    use futures::executor::block_on;
    use lambda_models::{get_firehose_input_example, StationTypes, WeatherDataInput};

    fn generate_event(event_type: StationTypes) -> KinesisFirehoseEventRecord {
        let mut station: WeatherDataInput =
            serde_json::from_str(get_firehose_input_example().as_str()).unwrap();
        station.station_type = event_type;
        let data = serde_json::to_vec(&station).unwrap();
        KinesisFirehoseEventRecord {
            record_id: None,
            approximate_arrival_timestamp: MillisecondTimestamp(chrono::DateTime::from_utc(
                NaiveDateTime::from_timestamp(1_000_000_000, 0),
                Utc,
            )),
            data: Base64Data(data),
        }
    }

    #[test]
    pub fn process_authorized_event() {
        let input = generate_event(StationTypes::AUTHORIZED);
        let output = block_on(process_input(input)).unwrap();
        println!("{:#?}", output);
    }
}
