use crate::LambdaError;
use aws_lambda_events::event::sns::{SnsEntity, SnsEvent};
use lambda_models::WeatherDataOutput;
use rusoto_core::Region;
use rusoto_s3::{PutObjectRequest, S3};
use std::str::FromStr;

/// Main data processing point.
///
/// This function proceed SNS input.
/// As input we need to provide common SNS event model.
/// It have several records to process and if processing
/// is success this function will return nothing.
///
/// # Errors
///
/// Will return `LambdaError` if any problem occur during processing.
/// * `S3_BUCKET` env variable was not provided;
/// * `AWS_REGION` env variable was not found;
/// * `AWS_REGION` env variable is incorrect;
pub async fn process(input: SnsEvent) -> Result<(), LambdaError> {
    info!("Lambda ignition");
    let current_region = std::env::var("AWS_REGION")?;
    let s3_bucket = std::env::var("S3_BUCKET")?;
    let region = Region::from_str(current_region.as_str())?;
    for record in input.records {
        debug!("Got event from {:#?}", record.event_source);
        debug!("Event version: {:#?}", record.event_version);
        match process_record(record.sns, &region, s3_bucket.as_str()).await {
            Ok(_) => info!("Message processed successfully"),
            Err(err) => warn!("Can't process message. Cause: {}", err.to_string()),
        }
    }
    Ok(())
}

/// This function is actual single `SNS-message` processor
/// It validates inputs and send it to S3 bucket.
///
/// # Errors
///
/// Function may fail in several cases:
/// * Message can't be parsed to [`WeatherDataOutput`]
/// * Publishing to S3 failed for some reason.
pub async fn process_record(
    record: SnsEntity,
    region: &Region,
    s3_bucket: &str,
) -> Result<(), LambdaError> {
    debug!(
        "Type: {:#?}",
        record.type_.unwrap_or_else(|| String::from(""))
    );
    match record.message {
        None => {
            error!("No message was found");
            Ok(())
        }
        Some(message_json) => {
            let s3_client = rusoto_s3::S3Client::new(region.clone());
            let weather_data: WeatherDataOutput = serde_json::from_str(message_json.as_str())?;
            let mut request = PutObjectRequest::default();
            request.bucket = String::from(s3_bucket);
            request.body = Some(message_json.into_bytes().into());
            request.key = format!(
                "UNAUTHORIZED/{}/{}/{}.txt",
                weather_data.country,
                weather_data.city,
                uuid::Uuid::new_v4().to_string()
            );
            s3_client.put_object(request).await?;
            Ok(())
        }
    }
}
