#[macro_use]
extern crate log;
extern crate tokio;

use fern::colors;
use log::LevelFilter;
use std::error::Error;
use std::str::FromStr;

mod processor;

type AppError = Box<dyn Error + Send + Sync + 'static>;

/// Setting up logger at startup
fn setup_logger() -> Result<(), AppError> {
    let log_level = std::env::var("LOG_LEVEL").unwrap_or_else(|_| String::from("INFO"));
    let log_file =
        std::env::var("LOG_FILE").unwrap_or_else(|_| String::from("/var/log/sqs_listener.log"));

    let colors = colors::ColoredLevelConfig::new()
        .info(colors::Color::Green)
        .warn(colors::Color::Magenta)
        .debug(colors::Color::Blue)
        .error(colors::Color::Red)
        .trace(colors::Color::BrightRed);
    fern::Dispatch::new()
        .format(move |out, message, record| {
            out.finish(format_args!(
                "[{}] {}",
                colors.color(record.level()),
                message
            ))
        })
        .level(LevelFilter::from_str(log_level.as_str()).unwrap())
        .chain(std::io::stdout())
        .chain(fern::log_file(log_file)?)
        .apply()?;
    Ok(())
}

/// Main execution point.
/// In this function we setup connection
/// and start listening to SQS
///
/// # Errors
/// `SQS_QUEUE_ID` env was not provided
/// `AWS_REGION` env was not provided
#[tokio::main]
async fn main() -> Result<(), AppError> {
    dotenv::dotenv().ok();
    setup_logger()?;

    processor::process().await?;
    Ok(())
}
