use crate::AppError;
use lambda_models::WeatherDataOutput;
use rusoto_core::Region;
use rusoto_s3::{GetObjectRequest, S3};
use rusoto_sns::{PublishInput, Sns, SnsClient};
use rusoto_sqs::{DeleteMessageRequest, Message, ReceiveMessageRequest, Sqs, SqsClient};
use std::str::FromStr;
use std::time::Duration;
use tokio::fs::File;
use tokio::prelude::*;
use tokio::time::delay_for;

fn build_error(description: String) -> std::io::Error {
    std::io::Error::new(std::io::ErrorKind::Other, description)
}

/// ## Main processing function
/// This function initializing `SQSClient`
/// and start pooling for new messages.
///
/// # Errors
/// * `SQS_QUEUE_ID` env was not found.
/// * `AWS_REGION` env was not found.
/// * `S3_BUCKET` env was not found.
/// * `AWS_REGION` is incorrect (Unknown AWS region).
pub async fn process() -> Result<(), AppError> {
    let sqs_queue_id = std::env::var("SQS_QUEUE_ID")
        .map_err(|_| build_error(String::from("SQS_QUEUE_ID was not provided")))?;
    debug!("Found SQS queue ID: {}", sqs_queue_id.as_str());
    let aws_region = std::env::var("AWS_REGION")
        .map_err(|_| build_error(String::from("AWS_REGION was not provided")))?;
    debug!("Found AWS region: {}", aws_region.as_str());
    let s3_bucket = std::env::var("S3_BUCKET")
        .map_err(|_| build_error(String::from("S3_BUCKET was not provided")))?;
    debug!("S3 BUCKET: {}", s3_bucket.as_str());
    let sns_topic = std::env::var("SNS_TOPIC")
        .map_err(|_| build_error(String::from("SNS_TOPIC was not provided")))?;
    debug!("SNS TOPIC: {}", s3_bucket.as_str());

    let aws_region = Region::from_str(aws_region.as_str())
        .map_err(|_| build_error(String::from("Incorrect region provided")))?;

    pool_messages(sqs_queue_id, s3_bucket, sns_topic, aws_region).await;
    Ok(())
}

/// Request pending message from sqs
/// and process them.
/// All errors handled here to continue execution in any case.
/// # Parameters
/// * `sqs_queue_id`: `SQS` queue to poll for messages;
/// * `s3_bucket`: `S3` bucket to get files described in incoming messages;
/// * `sns_topic`: `SNS` topic to publish matched weather data;
/// * `aws_region`: current `AWS` region.
async fn pool_messages(
    sqs_queue_id: String,
    s3_bucket: String,
    sns_topic: String,
    aws_region: Region,
) {
    info!("Start pooling message from SQS");
    loop {
        info!("Sleeping for 2 seconds");
        delay_for(Duration::from_secs(2)).await;
        let sqs_client = SqsClient::new(aws_region.clone());
        let messages = request_messages(&sqs_client, sqs_queue_id.as_str()).await;
        if messages.is_err() {
            warn!(
                "Can't receive messages. Cause: {}",
                messages.err().unwrap().to_string()
            );
            continue;
        }
        for message in &messages.unwrap() {
            let message = message.clone();
            let message_id = message
                .message_id
                .clone()
                .unwrap_or_else(|| String::from("with unknown ID"));
            let processing_res = process_message(
                s3_bucket.as_str(),
                sns_topic.as_str(),
                &aws_region,
                message.clone(),
            )
            .await;
            match processing_res {
                Ok(_) => {
                    info!("Message {} processed successfully", message_id);
                    if let Err(error) =
                        delete_message(&sqs_client, sqs_queue_id.as_str(), message.clone()).await
                    {
                        warn!("Can't delete message from queue. {}", error.to_string());
                    };
                }
                Err(error) => error!(
                    "Can't process message: {:?}. Cause: {}",
                    message.message_id,
                    error.to_string()
                ),
            }
        }
    }
}

/// Delete message from `SQS` queue.
///
/// # Parameters
/// * `sqs_client`: Connection to `SQS`;
/// * `queue_id`: Link to queue;
/// * `message`: Message from SQS to delete.
///
/// # Errors
///
/// * Message receipt empty;
/// * Can't delete message from queue.
async fn delete_message(
    sqs_client: &SqsClient,
    queue_id: &str,
    message: Message,
) -> Result<(), AppError> {
    let mut delete_request = DeleteMessageRequest::default();
    delete_request.queue_url = String::from(queue_id);
    if message.receipt_handle.is_none() {
        return Err(Box::new(build_error(String::from(
            "Unknown receipt handle in message",
        ))));
    }
    delete_request.receipt_handle = String::from(message.receipt_handle.unwrap().as_str());
    sqs_client.delete_message(delete_request).await?;
    Ok(())
}

/// Request new messages from SQS queue
///
/// # Parameters
/// * `sqs_client`: Connection to `SQS`;
/// * `queue_id`: Link to queue;
///
/// # Errors
/// * Can't receive messages;
/// * The `Queue` is empty.
async fn request_messages(
    sqs_client: &SqsClient,
    queue_id: &str,
) -> Result<Vec<Message>, AppError> {
    let mut request = ReceiveMessageRequest::default();
    request.queue_url = String::from(queue_id);
    request.max_number_of_messages = Some(10);
    let sqs_result = sqs_client.receive_message(request).await.map_err(|err| {
        Box::new(build_error(format!(
            "Can't request more messages. Cause: {}",
            err.to_string()
        )))
    })?;
    if sqs_result.messages.is_none() {
        return Err(Box::new(build_error(String::from("The queue is empty"))));
    }
    Ok(sqs_result.messages.unwrap())
}

/// Get message body and delete it from queue.
///
/// # Parameters
/// * `s3_bucket`: `S3` bucket where parquet is located;
/// * `sns_topic`: `SNS` topic to publish matched weather data;
/// * `aws_region`: Current AWS region;
/// * `message`: `Message` to process.
///
/// # Errors
/// * Can't load provided file.
async fn process_message(
    s3_bucket: &str,
    sns_topic: &str,
    aws_region: &Region,
    message: Message,
) -> Result<(), AppError> {
    debug!("Processing message");
    if message.body.is_none() {
        return Ok(());
    }
    let body = message.body.unwrap().clone();
    debug!("Received s3 file key: {}", body.as_str());
    let uuid = uuid::Uuid::new_v4().to_string();
    let file_name = format!("/tmp/{}.parquet", uuid);
    load_and_save_parquet(body, aws_region, String::from(s3_bucket), file_name.clone()).await?;
    parse_and_publish_parquet(file_name, sns_topic, aws_region).await?;
    Ok(())
}

/// Loading and saving parquet file to filesystem.
/// All saved parquet files available until scaling down.
///
/// # Parameters
/// * `file_key`: key to file in s3 storage;
/// * `aws_region`: Current AWS region where s3 located;
/// * `s3_bucket`: bucket obviously;
/// * `file_name`: name of file on local system where to store parquet.
///
/// # Errors
/// * Can't load object from S3;
/// * Can't read object bytes;
/// * Can't create file in filesystem;
/// * Can't write to created file.
async fn load_and_save_parquet(
    file_key: String,
    aws_region: &Region,
    s3_bucket: String,
    file_name: String,
) -> Result<(), AppError> {
    debug!("Loading s3 object");
    let s3_client = rusoto_s3::S3Client::new(aws_region.clone());
    let mut object_request = GetObjectRequest::default();
    object_request.bucket = s3_bucket.clone();
    object_request.key = file_key.clone();
    let s3_object = s3_client.get_object(object_request).await?;
    if s3_object.body.is_none() {
        return Err(Box::new(build_error(String::from(
            "S3 object body is null",
        ))));
    }
    let body = s3_object.body.unwrap();
    let mut contents: Vec<u8> = vec![];
    body.into_async_read().read_to_end(&mut contents).await?;
    let mut file = File::create(file_name).await?;
    file.write_all(contents.as_slice()).await?;
    Ok(())
}

/// Parse parquet file and publish it to SNS.
/// # Parameters:
/// * `filename`: file in local system to read parquet file;
/// * `aws_region`: current AWS region;
///
/// # Errors
/// * The humidity is above 10;
/// * The pressure is above 100;
/// * The temperature is less than 30.
async fn parse_and_publish_parquet(
    file_name: String,
    sns_topic: &str,
    aws_region: &Region,
) -> Result<(), AppError> {
    debug!("Parsing parquet file {}", file_name.as_str());
    let records = lambda_models::WeatherDataOutput::from_parquet_file(file_name).await?;
    let sns_client = rusoto_sns::SnsClient::new(aws_region.clone());
    for item in records {
        debug!("{:#?}", item.clone());
        if item.humidity > 10 && item.pressure > 100 && item.temperature < 30 {
            match publish_to_sns(item, sns_topic, &sns_client).await {
                Ok(_) => {
                    info!("Weather data successfully processed");
                }
                Err(err) => {
                    warn!("Can't send weather data to SNS. Cause {}", err.to_string());
                }
            }
        } else {
            warn!("Weather data is ok");
        }
    }
    Ok(())
}

/// Publish valid messages to SNS
///
/// # Parameters
/// `weather_data`: parsed parquet from `S3`;
/// `sns_topic`: topic to publish to;
/// `sns_client`: Connection to `SNS`.
///
/// # Errors
/// * Can't serialize `WeatherData` to JSON;
/// * Can't publish to `SNS`.
async fn publish_to_sns(
    weather_data: WeatherDataOutput,
    sns_topic: &str,
    sns_client: &SnsClient,
) -> Result<(), AppError> {
    debug!("Publishing to SNS");
    let mut publish_input = PublishInput::default();
    publish_input.message = serde_json::to_string(&weather_data)?;
    publish_input.topic_arn = Some(String::from(sns_topic));
    sns_client.publish(publish_input).await?;
    Ok(())
}
