use crate::LambdaError;
use aws_lambda_events::event::s3::{S3Event, S3EventRecord};
use rusoto_core::Region;
use rusoto_sqs::{SendMessageRequest, Sqs};
use std::io::ErrorKind;
use std::str::FromStr;

/// Main data processing point.
///
/// This function proceed S3 put events.
/// After object was created in s3 this processor will send information
/// about target records in sqs.
///
/// # Errors
///
/// Will return `Err` if any problem occur during processing.
pub async fn process(event: S3Event) -> Result<(), Vec<LambdaError>> {
    info!("Handler started");
    for record in event.records {
        match process_input(record).await {
            Ok(_) => info!("Information successfully sent."),
            Err(err) => {
                error!("Can't send information about this event");
                error!("{}", err.to_string())
            }
        }
    }
    Ok(())
}

/// Main processor logic.
/// It will publish information about new object from s3 in SQS.
///
/// `AWS_REGION` is variable provided by AWS lambda environment.
/// `SQS_QUEUE_ID` is injected by terraform.
///
/// # Errors
/// * The `AWS_REGION` variable not found;
/// * The `AWS_REGION` env variable is incorrect;
/// * The `SQS_QUEUE_ID` env is not set;
/// * The `SQS_QUEUE_ID` env is incorrect.
pub async fn process_input(record: S3EventRecord) -> Result<(), LambdaError> {
    let queue_url = std::env::var("SQS_QUEUE_ID")?;
    let current_region = std::env::var("AWS_REGION")?;
    let client = rusoto_sqs::SqsClient::new(Region::from_str(current_region.as_str())?);
    info!("Created sqs client");
    let mut message = SendMessageRequest::default();
    info!("Resolved queue url: {}", queue_url);
    message.queue_url = queue_url;
    let object_key = record.s3.object.key;
    info!("Event data key: {:?}", object_key);
    if object_key.is_none() {
        return Err(Box::new(std::io::Error::new(
            ErrorKind::Other,
            String::from("Object path is unknown!"),
        )));
    }
    message.message_body = object_key.unwrap();
    client.send_message(message).await?;
    Ok(())
}
